// Copyright 2022 Memgraph Ltd.
//
// Use of this software is governed by the Business Source License
// included in the file licenses/BSL.txt; by using this file, you agree to be bound by the terms of the Business Source
// License, and you may not use this file except in compliance with the Business Source License.
//
// As of the Change Date specified in that file, in accordance with
// the Business Source License, use of this software will be governed
// by the Apache License, Version 2.0, included in the file
// licenses/APL.txt.

#pragma once

#include <memory>
#include <optional>
#include <unordered_map>
#include <utility>
#include <vector>

#include "common/types.hpp"
#include "storage/v2/id_types.hpp"
#include "storage/v2/indices.hpp"
#include "storage/v2/property_value.hpp"
#include "storage/v2/temporal.hpp"
#include "storage/v2/transaction.hpp"
#include "storage/v2/vertex.hpp"
#include "utils/result.hpp"

namespace memgraph::storage {

class SchemaViolationException : public utils::BasicException {
  using utils::BasicException::BasicException;
};

struct SchemaProperty {
  common::SchemaPropertyType type;
  PropertyId property_id;
};

struct SchemaViolation {
  enum class ValidationStatus : uint8_t {
    VERTEX_HAS_NO_PRIMARY_LABEL,
    VERTEX_HAS_NO_PROPERTY,
    NO_SCHEMA_DEFINED_FOR_LABEL,
    VERTEX_PROPERTY_WRONG_TYPE
  };

  SchemaViolation(ValidationStatus status, LabelId label);

  SchemaViolation(ValidationStatus status, LabelId label, SchemaProperty violated_type);

  SchemaViolation(ValidationStatus status, LabelId label, SchemaProperty violated_type,
                  PropertyValue violated_property_value);

  ValidationStatus status;
  LabelId label;
  std::optional<SchemaProperty> violated_type;
  std::optional<PropertyValue> violated_property_value;
};

/// Structure that represents a collection of schemas
/// Schema can be mapped under only one label => primary label
class Schemas {
 public:
  using Schema = std::pair<LabelId, std::vector<SchemaProperty>>;
  using SchemasMap = std::unordered_map<LabelId, std::vector<SchemaProperty>>;
  using SchemasList = std::vector<Schema>;

  Schemas() = default;
  Schemas(const Schemas &) = delete;
  Schemas(Schemas &&) = delete;
  Schemas &operator=(const Schemas &) = delete;
  Schemas &operator=(Schemas &&) = delete;
  ~Schemas() = default;

  [[nodiscard]] SchemasList ListSchemas() const;

  [[nodiscard]] SchemasList GetSchema(LabelId primary_label) const;

  [[nodiscard]] bool CreateSchema(LabelId label, const std::vector<SchemaProperty> &schemas_types);

  [[nodiscard]] bool DeleteSchema(LabelId label);

  [[nodiscard]] std::optional<SchemaViolation> ValidateVertex(LabelId primary_label, const Vertex &vertex);

 private:
  SchemasMap schemas_;
};

inline std::optional<common::SchemaPropertyType> PropertyTypeToSchemaType(const PropertyValue &property_value) {
  switch (property_value.type()) {
    case PropertyValue::Type::Bool: {
      return common::SchemaPropertyType::BOOL;
    }
    case PropertyValue::Type::Int: {
      return common::SchemaPropertyType::INT;
    }
    case PropertyValue::Type::Double: {
      return common::SchemaPropertyType::FLOAT;
    }
    case PropertyValue::Type::String: {
      return common::SchemaPropertyType::STRING;
    }
    case PropertyValue::Type::TemporalData: {
      switch (property_value.ValueTemporalData().type) {
        case TemporalType::Date: {
          return common::SchemaPropertyType::DATE;
        }
        case TemporalType::LocalDateTime: {
          return common::SchemaPropertyType::LOCALDATETIME;
        }
        case TemporalType::LocalTime: {
          return common::SchemaPropertyType::LOCALTIME;
        }
        case TemporalType::Duration: {
          return common::SchemaPropertyType::DURATION;
        }
      }
    }
    case PropertyValue::Type::Null:
    case PropertyValue::Type::Map:
    case PropertyValue::Type::List: {
      return std::nullopt;
    }
  }
}

inline std::string SchemaPropertyToString(const common::SchemaPropertyType type) {
  switch (type) {
    case common::SchemaPropertyType::BOOL: {
      return "Bool";
    }
    case common::SchemaPropertyType::INT: {
      return "Integer";
    }
    case common::SchemaPropertyType::FLOAT: {
      return "Float";
    }
    case common::SchemaPropertyType::STRING: {
      return "String";
    }
    case common::SchemaPropertyType::DATE: {
      return "Date";
    }
    case common::SchemaPropertyType::LOCALTIME: {
      return "LocalTime";
    }
    case common::SchemaPropertyType::LOCALDATETIME: {
      return "LocalDateTime";
    }
    case common::SchemaPropertyType::DURATION: {
      return "Duration";
    }
  }
}

}  // namespace memgraph::storage
